# Purpose: Plot data
# Author: Minzhe Hu
# Date: 2025.1.6
# Email: hmz2018@mail.ustc.edu.cn
import numpy as np
import matplotlib.pyplot as plt
from collections.abc import Sequence


def plot(data: np.ndarray, dx=None, fs=None, ax=None, obj='waveform', dpi=300,
         title=None, transpose=False, t0=0, x0=0, pick=None, f=None, k=None,
         t=None, c=None, cmap=None, vmin=None, vmax=None, dB=False,
         xmode='distance', tmode='time', xlim=None, ylim=None, xlog=False,
         ylog=False, xinv=False, yinv=False, xlabel=True, ylabel=True,
         xticklabels=True, yticklabels=True, colorbar=True, colorbar_label=None,
         savefig=None):
    """
    Plot several types of 2-D seismological data.

    :param data: numpy.ndarray. Data to plot.
    :param dx: Channel interval in m.
    :param fs: Sampling rate in Hz.
    :param ax: Matplotlib.axes.Axes or tuple. Axes to plot. A tuple for new
        figsize. If not specified, the function will directly display the image
        using matplotlib.pyplot.show().
    :param obj: str. Type of data to plot. It should be one of 'waveform',
        'phasepick', 'spectrum', 'spectrogram', 'fk', or 'dispersion'.
    :param dpi: int. The resolution of the figure in dots-per-inch.
    :param title: str. The title of this axes.
    :param transpose: bool. Transpose the figure or not.
    :param t0, x0: The beginning of time and space.
    :param pick: dictionary of sequence of picked phases. Key should be 'P' for
        P phase, 'S' for S phase and 'N' for unknown phase type. Required if
        obj=='phasepick'.
    :param f: Sequence of frequency. Required if obj is one of 'spectrum',
        'spectrogram', 'fk' or 'dispersion'.
    :param k: Wavenumber sequence. Required if obj=='fk'.
    :param t: Time sequence. Required if obj=='spectrogram'.
    :param c: Phase velocity sequence. Required if obj=='dispersion'.
    :param cmap: str or Colormap. The Colormap instance or registered colormap
        name used to map scalar data to colors.
    :param vmin, vmax: Define the data range that the colormap covers.
    :param dB: bool. Transfer data unit to dB and take 1 as the reference value.
    :param xmode: str. 'distance' or 'channel'.
    :param tmode: str. 'time' or 'sampling'.
    :param xlim, ylim: Set the x-axis and y-axis view limits.
    :param xlog, ylog: bool. If True, set the x-axis' or y-axis' scale as log.
    :param xlabel, yinv: bool. If True, invert x-axis or y-axis.
    :param xlabel, ylabel: bool or str. Whether to plot a label or what label to
        plot for x-axis or y-axis.
    :param xticklabels, yticklabels: bool or sequence of str. Whether to plot
        ticklabels or what ticklabels to plot for x-axis or y-axis.
    :param colorbar: bool, str or Matplotlib.axes.Axes. Bool means plot colorbar
        or not. Str means the location of colorbar. Axes means the Axes into
        which the colorbar will be drawn.
    :param savefig: str or bool. Figure name to save if needed. If True,
        it will be set to parameter obj.
    """
    nch, nt = data.shape
    if ax is None:
        ax = (6, 5)
    if isinstance(ax, tuple):
        fig, ax = plt.subplots(1, figsize=ax, dpi=dpi)
        show = True
    else:
        show = False

    if obj in ['waveform', 'phasepick']:
        cmap = 'RdBu' if cmap is None else cmap
        vmax = np.percentile(abs(data), 80) if vmax is None else vmax
        vmin = -vmax if vmin is None else vmin
        origin = 'upper'
        if fs is None or tmode == 'sampling':
            ylabel_default = 'Sampling points'
            fs = 1
        elif tmode == 'time':
            ylabel_default = 'Time (s)'

        if dx is None or xmode.lower() == 'channel':
            xlabel_default = 'Channel'
            extent = [x0, x0 + nch, t0 + nt / fs, t0]
        elif xmode.lower() == 'distance':
            xlabel_default = 'Disitance (km)'
            extent = [x0 * 1e-3, (x0 + nch * dx) * 1e-3, t0 + nt / fs, t0]

        if obj == 'phasepick' and len(pick):
            pick_color = {'P': 'r', 'S': 'b', 'N': 'k'}
            for phase, pck in pick.items():
                if len(pck):
                    pck = np.array(pck).astype(float)
                    if xmode.lower() == 'distance':
                        pck[:, 0] = (x0 + pck[:, 0] * dx) * 1e-3
                    elif xmode.lower() == 'channel':
                        pck[:, 0] = x0 + pck[:, 0]
                    if tmode.lower() == 'sampling':
                        pck[:, 1] = pck[:, 1] / fs
                    ax.scatter(pck[:,0], t0 + pck[:,1], marker=',', s=0.1,
                            c=pick_color[phase])

    elif obj in ['spectrum', 'spectrogram', 'fk', 'dispersion']:
        if isinstance(data[0,0], (complex, np.complex64)):
            data = abs(data)
        if dB:
            data = 20 * np.log10(data)
        cmap = 'jet' if cmap is None else cmap
        vmax = np.percentile(abs(data), 80) if vmax is None else vmax
        vmin = np.percentile(abs(data), 20) if vmin is None else vmin
        if obj == 'spectrum':
            origin = 'lower'
            if dx is None or xmode.lower() == 'channel':
                xlabel_default = 'Channel'
                extent = [x0, x0 + nch, min(f), max(f)]
            elif xmode.lower() == 'distance':
                xlabel_default = 'Disitance (km)'
                extent = [x0 * 1e-3, (x0 + nch * dx) * 1e-3, min(f), max(f)]
            ylabel_default = 'Frequency (Hz)'
        elif obj == 'spectrogram':
            data = data.T
            origin = 'lower'
            xlabel_default = 'Time (s)'
            ylabel_default = 'Frequency (Hz)'
            extent = [t0 + min(t), t0 + max(t), min(f), max(f)]
        elif obj == 'fk':
            origin = 'lower'
            xlabel_default = 'Wavenumber (m$^{-1}$)'
            ylabel_default = 'Frequency (Hz)'
            extent = [min(k), max(k), min(f), max(f)]
        elif obj == 'dispersion':
            data = data.T
            origin = 'lower'
            xlabel_default = 'Frequency (Hz)'
            ylabel_default = 'Velocity (m/s)'
            extent = [min(f), max(f), min(c), max(c)]

    if transpose:
        if origin == 'lower':
            extent = [extent[2], extent[3], extent[0], extent[1]]
        else:
            origin = 'lower'
            extent = [extent[3], extent[2], extent[0], extent[1]]
        (xlabel_default, ylabel_default) = (ylabel_default, xlabel_default)
        data = data.T

    xlabel = xlabel if isinstance(xlabel, str) else \
            xlabel_default if xlabel else None
    ylabel = ylabel if isinstance(ylabel, str) else \
            ylabel_default if ylabel else None

    bar = ax.imshow(data.T, vmin=vmin, vmax=vmax, extent=extent, aspect='auto',
                    origin=origin, cmap=cmap)
    if title:
        ax.set_title(title)
    ax.set_xlabel(xlabel)
    ax.set_ylabel(ylabel)
    if isinstance(xticklabels, Sequence):
        ax.set_xticklabels(xticklabels)
    elif not xticklabels:
        ax.set_xticklabels([])
    
    if isinstance(yticklabels, Sequence):
        ax.set_yticklabels(yticklabels)
    elif not yticklabels:
        ax.set_yticklabels([])
    if xinv:
        ax.invert_xaxis()
    if yinv:
        ax.invert_yaxis()
    if ylim:
        ax.set_ylim(ylim)
    if xlim:
        ax.set_xlim(xlim)
    if xlog:
        ax.set_xscale('log')
    if ylog:
        ax.set_yscale('log')
    if colorbar:
        if colorbar is True:
            cbar = plt.colorbar(bar, ax=ax, location='right')
        elif isinstance(colorbar, str):
            cbar = plt.colorbar(bar, ax=ax, location=colorbar)
        else:
            cbar = plt.colorbar(bar, cax=colorbar)
        if colorbar_label is not None:
            cbar.set_label(colorbar_label)

    if savefig:
        if not isinstance(savefig, str):
            savefig = obj + '.png'
        plt.tight_layout()
        plt.savefig(savefig)
        plt.close()
    elif show:
        plt.show()
    else:
        return ax



from setuptools import setup, find_packages


setup(
    name='DASPy-toolbox', version='1.1.5',
    description=(
        'DASPy is an open-source project dedicated to provide a python package '
        'for DAS (Distributed Acoustic Sensing) data processing, which '
        'comprises classic seismic data processing techniques and Specialized '
        'algorithms for DAS applications.'
        ),
    long_description=open('README.md').read(),
    author='Minzhe Hu, Zefeng Li',
    author_email='hmz2018@mail.ustc.edu.cn',
    maintainer='Minzhe Hu',
    maintainer_email='hmz2018@mail.ustc.edu.cn',
    license='MIT License',
    url='https://github.com/HMZ-03/DASPy',
    packages=find_packages(),
    entry_points={  
        'console_scripts': [
            'daspy = daspy.main:main',
            ]
    },
    include_package_data=True,
    package_data={
        'daspy': ['core/example.pkl']
    },
    classifiers=[
        'Operating System :: OS Independent',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3'
    ],
    python_requires='>=3.9',
    install_requires=[
        'numpy',
        'scipy>=1.13',
        'matplotlib',
        'geographiclib',
        'pyproj',
        'h5py',
        'segyio',
        'nptdms',
        'tqdm'
    ]
)



# Purpose: Module for handling Section objects.
# Author: Minzhe Hu
# Date: 2025.4.2
# Email: hmz2018@mail.ustc.edu.cn
import warnings
import os
import numpy as np
from copy import deepcopy
from typing import Iterable
from datetime import datetime
from daspy.core.dasdatetime import DASDateTime, utc
from daspy.core.write import write
from daspy.basic_tools.visualization import plot
from daspy.basic_tools.preprocessing import (phase2strain, normalization,
                                             demeaning, detrending, stacking,
                                             cosine_taper, downsampling,
                                             padding, trimming,
                                             time_integration,
                                             time_differential,
                                             distance_integration)
from daspy.basic_tools.filter import (bandpass, bandstop, lowpass,
                                      lowpass_cheby_2, highpass, envelope)
from daspy.basic_tools.freqattributes import (spectrum, spectrogram,
                                              fk_transform)
from daspy.advanced_tools.channel import channel_checking, turning_points
from daspy.advanced_tools.denoising import (curvelet_denoising,
                                            common_mode_noise_removal,
                                            spike_removal)
from daspy.advanced_tools.decomposition import fk_filter, curvelet_windowing
from daspy.advanced_tools.strain2vel import (slant_stacking, fk_rescaling,
                                             curvelet_conversion)


class Section(object):
    def __init__(self, data, dx, fs, start_channel=0, start_distance=0,
                 start_time=0, **kwargs):
        """
        :param data: numpy.ndarray. Data recorded by DAS.
        :param dx: number. Channel interval in m.
        :param fs: number. Sampling rate in Hz.
        :param start_channel: int. Channel number of the first channel.
        :param start_distance: number. Distance of the first channel, in m.
        :param start_time: number or DASDateTime. Time of the first
            sampling point. If number, the unit is s.
        :param origin_time: number or DASDateTime. Ocurance time of the event.
        :param gauge_length: number. Gauge length in m.
        :param data_type: str. Can be 'phase shift', 'phase change rate',
            'strain', 'strain rate', 'displacement', 'velocity', 'acceleration',
            or normalized above parameters.
        :param scale: number. Scale or gain of data.
        :param geometry: numpy.ndarray. Should include latitude and longitude
            (first two columns), and can also include depth (last column).
        :param turning_channels: sequnce of channel numbers. Channel numbers of
            turning points.
        :param headers: dict. Other headers.
        :param source: str or pathlib.PosixPath. Path to the source file.
        :param source_type: str. Raw type it read from.
        """
        if data.ndim == 1:
            data = data[np.newaxis, :]
        self.data = data
        self.dx = dx
        self.fs = fs
        self.start_channel = start_channel
        self.start_distance = start_distance
        self.start_time = start_time
        opt_attrs = ['origin_time', 'gauge_length', 'data_type', 'scale',
                     'geometry', 'turning_channels', 'headers', 'source',
                     'source_type']
        for attr in opt_attrs:
            if attr in kwargs:
                setattr(self, attr, kwargs.pop(attr))

    def __str__(self):
        n = max(map(len, self.__dict__.keys()))
        describe = '{}: shape{}\n'.format('data'.rjust(n), self.data.shape)
        for key, value in self.__dict__.items():
            if key == 'data':
                continue
            if key == 'geometry':
                describe += '{}: shape{}\n'.format(key.rjust(n), value.shape)
            elif key in ['dx', 'start_distance', 'gauge_length']:
                describe += '{}: {} m\n'.format(key.rjust(n), value)
            elif key == 'fs':
                describe += '{}: {} Hz\n'.format(key.rjust(n), value)
            elif key == 'start_time':
                if isinstance(value, DASDateTime):
                    describe += '{}: {}\n'.format(key.rjust(n), value)
                else:
                    describe += '{}: {} s\n'.format(key.rjust(n), value)
            else:
                describe += '{}: {}\n'.format(key.rjust(n), value)
        return describe

    __repr__ = __str__

    def __add__(self, other):
        """
        Join two sections in time.
        """
        out = self.copy()
        if isinstance(other, Section):
            if other.dx != self.dx:
                if self.dx is None:
                    out.dx = other.dx
                elif other.dx is not None:
                    raise ValueError('These two Sections have different dx, '
                                     'please check.')
            if other.fs != self.fs:
                if self.fs is None:
                    out.fs = other.fs
                elif other.fs is not None:
                    raise ValueError('These two Sections have different fs, '
                                     'please check.')
            if isinstance(self.start_time, DASDateTime) and \
                isinstance(other.start_time, DASDateTime):
                if abs(other.start_time - self.end_time) > 0.1:
                    if abs(other.end_time - self.start_time) <= 0.1:
                        warnings.warn('According to the time information of the'
                                      ' two Sections, the order of addition is '
                                      'reversed.')
                        return other + self
                    else:
                        warnings.warn('The start time of the second Section '
                                      f'({other.start_time}) is inconsistent '
                                      'with the end time of the first Section ('
                                      f'{self.end_time}).')
            data = other.data
        elif isinstance(other, np.ndarray):
            data = other
        elif isinstance(other, list):
            data = np.array(other)
        else:
            raise TypeError('The input should be Section or np.ndarray.')

        if len(data) != self.nch:
            if len(data[0]) == self.nch:
                data = data.T
            else:
                raise ValueError('These two Sections have different number of '
                                 'channels, please check.')
        if out.data is None:
            out.data = data
        else:
            out.data = np.hstack((out.data, data))

        return out

    @property
    def shape(self):
        return self.data.shape

    @property
    def dt(self):
        return 1 / self.fs

    @property
    def nch(self):
        return len(self.data)

    @property
    def nt(self):
        return self.data.shape[1]

    @property
    def end_channel(self):
        return self.start_channel + self.nch - 1

    @property
    def distance(self):
        return self.nch * self.dx
    
    @property
    def channel_distance(self):
        return self.dx * np.arange(self.nch) + self.start_channel

    @property
    def end_distance(self):
        return self.start_distance + self.nch * self.dx

    @property
    def duration(self):
        return self.nt / self.fs

    @property
    def end_time(self):
        return self.start_time + self.nt / self.fs

    def copy(self):
        return deepcopy(self)

    @classmethod
    def from_obspy_stream(cls, st, channel_no='auto'):
        """
        Construct a Section from a obspy.core.stream.Stream instance.

        :param patch: obspy.core.stream.Stream. An instance of
            obspy.core.stream.Stream for construction.
        :param channel_no: None or str. None for no channel number information,
            'channel', 'station' for use the channel or station information of
            each trace and 'auto' for automatically detect.
        """
        stime = min([tr.stats['starttime'] for tr in st])
        etime = max([tr.stats['endtime'] for tr in st])
        st.trim(starttime=stime, endtime=etime, pad=True, fill_value=np.nan)
        matadata = [(tr.stats['sampling_rate'], tr.stats['delta'],
                     tr.stats['npts'], tr.stats['calib']) for tr in st]
        assert len(set(matadata)) == 1, ('The metadata of all traces in the '
                                         'stream should be the same.')
        nch = len(st)
        nt = st[0].stats.npts
        fs = st[0].stats.sampling_rate
        start_time = DASDateTime.from_datetime(st[0].stats.starttime.datetime).\
            replace(tzinfo=utc)
        scale = st[0].stats.calib
        source = type(st)
        data = np.zeros((nch, nt))

        if channel_no == 'channel':
            channel_no = np.array([int(tr.stats.channel) for tr in st])
        elif channel_no == 'station':
            channel_no = np.array([int(tr.stats.station) for tr in st])
        elif channel_no:
            if str.isdigit(st[0].stats.channel):
                channel_no = np.array([int(tr.stats.channel) for tr in st])
            elif str.isdigit(st[0].stats.station):
                channel_no = np.array([int(tr.stats.station) for tr in st])
        else:
            channel_no = np.arange(nch).astype(int)
        start_channel = min(channel_no)
        channel_no -= start_channel
        data = np.zeros((max(channel_no) + 1, nt))
        for i, tr in enumerate(st):
            data[channel_no[i]] = tr.data

        warnings.warn('obspy.core.stream.Stream doesn\'t include channel '
                      'interval. Please set dx manually.')
        return cls(data.astype(float), None, fs, start_channel=start_channel,
                   start_time=start_time, scale=scale, source=source)

    @classmethod
    def from_dascore_patch(cls, patch):
        """
        Construct a Section from a dascore.core.patch.Patch instance.

        :param patch: dascore.core.patch.Patch. An instance of
            dascore.core.patch.Patch for construction.
        :return: daspy.Section.
        """
        kwargs = {}
        if patch.dims == ('time', 'distance'):
            data = patch.data.T
            dx = patch.coords.coord_map['distance'].step
            kwargs['start_distance'] = patch.coords.coord_map['distance'].start
        elif patch.dims == ('time', 'channel'):
            data = patch.data.T
            dx = 1
            warnings.warn('This dascore.core.patch.Patch instance doesn\'t '
                          'include channel interval. Set dx to 1.')
            kwargs['start_channel'] = patch.coords.coord_map['channel'].start
        elif patch.dims == ('distance', 'time'):
            data = patch.data
            dx = patch.coords.coord_map['distance'].step
            kwargs['start_distance'] = patch.coords.coord_map['distance'].start
        elif patch.dim == ('channel', 'time'):
            data = patch.data
            dx = 1
            warnings.warn('This dascore.core.patch.Patch instance doesn\'t '
                          'include channel interval. Set dx to 1.')
            kwargs['start_channel'] = patch.coords.coord_map['channel'].start

        if isinstance(patch.coords.coord_map['time'].step, np.timedelta64):
            start_time = DASDateTime.fromtimestamp(
                patch.coords.coord_map['time'].start.item() / 1e9, tz=utc)
            fs = np.timedelta64(1, 's') / patch.coords.coord_map['time'].step
        else:
            start_time = patch.coords.coord_map['time'].start
            fs = patch.coords.coord_map['time'].step

        if hasattr(patch.attrs, 'gauge_length'):
            kwargs['gauge_length'] = patch.attrs.gauge_length
        if patch.attrs.data_type:
            kwargs['data_type'] = ' '.join(patch.attrs.data_type.split('_'))
        sec = cls(data, dx, fs, start_time=start_time,
                  headers=patch.attrs, source=type(patch), **kwargs)
        return sec

    @classmethod
    def from_lightguide_blast(cls, blast):
        """
        Construct a Section from a lightguide.blast.Blast instance.

        :param blast: lightguide.blast.Blast. An instance of
            lightguide.blast.Blast for construction.
        :return: daspy.Section.
        """
        sec = cls(blast.data, blast.channel_spacing, blast.sampling_rate,
                  start_time=DASDateTime.from_datetime(blast.start_time),
                  start_channel=blast.start_channel, data_type=blast.unit,
                  source=type(blast))
        return sec

    def to_obspy_stream(self):
        """
        Construct an instance of obspy.core.stream.Stream.

        :return: obspy.core.stream.Stream.
        """
        from obspy import Stream, Trace, UTCDateTime
        st = Stream()
        header = {'sampling_rate': self.fs}
        if not isinstance(datetime, self.start_time):
            warnings.warn('The type of start_time is not DASDateTime. The '
                          'starttime of Trace instances may be wrong')
            header['starttime'] = UTCDateTime(self.start_time)
        else:
            header['starttime'] = UTCDateTime(self.start_time.timestamp())
        if hasattr(self, 'scale'):
            header['calib'] = self.scale
        for i in range(self.nch):
            header_tr = deepcopy(header)
            header_tr['channel'] = str(self.start_channel + i)
            tr = Trace(self.data[i], header_tr)
            st += tr

        warnings.warn('obspy.core.stream.Stream doesn\'t include channel '
                      'interval.')
        return st

    def to_dascore_patch(self):
        """
        Construct an instance of dascore.core.patch.Patch.

        :return: dascore.core.patch.Patch.
        """
        from pint import Quantity
        from datetime import datetime
        from dascore.core import Patch, CoordManager
        from dascore.core.coords import CoordRange
        from dascore.utils.mapping import FrozenDict

        dims = ('time', 'distance')
        if isinstance(self.start_time, datetime):
            if self.start_time.tzinfo:
                stime = np.datetime64(self.start_time.astimezone(utc).
                                      replace(tzinfo=None))
                etime = np.datetime64(self.end_time.astimezone(utc).
                                      replace(tzinfo=None))
            else:
                stime = np.datetime64(self.start_time.replace(tzinfo=None))
                etime = np.datetime64(self.end_time.replace(tzinfo=None))

            time_range = CoordRange(units=(Quantity(1, 'second')),
                                    step=np.timedelta64(int(self.dt * 1e9),
                                                        'ns'),
                                    start=stime, stop=etime)
        else:
            stime = self.start_time
            etime = self.end_time
            time_range = CoordRange(units=(Quantity(1, 'second')), step=self.dt,
                                    start=stime, stop=etime)

        dist_range = CoordRange(units=(Quantity(1, 'meter')), step=self.dx,
                                start=self.start_distance,
                                stop=self.end_distance)
        coord_map = FrozenDict({'time': time_range, 'distance': dist_range})
        dim_map = FrozenDict({'time': ('time',), 'distance': ('distance', )})
        coords = CoordManager(dims=dims, coord_map=coord_map,
                              dim_map=dim_map)

        if self.source == Patch:
            attrs = self.headers
        else:
            from dascore.io.prodml.core import ProdMLPatchAttrs
            attrs = ProdMLPatchAttrs(coords=coords)
            kwargs = {}
            if hasattr(self, 'data_type'):
                kwargs['data_type'] = '_'.join(self.data_type.split(' '))
            if hasattr(self, 'gauge_length'):
                kwargs['gauge_length'] = self.gauge_length
                kwargs['gauge_length_units'] = Quantity(1, 'meter')

        return Patch(self.data.T, coords, dims, attrs)

    def to_lightguide_blast(self):
        """
        Construct an instance of lightguide.blast.Blast

        :return: lightguide.blast.Blast.
        """
        from lightguide.blast import Blast
        if not isinstance(self.start_time, datetime):
            warnings.warn('The type of start_time is not DASDateTime. The '
                          'starttime of Trace instances may be wrong')
            start_time = datetime.fromtimestamp(self.start_time)
        elif isinstance(self.start_time, DASDateTime):
            start_time = self.start_time.to_datetime()
        if start_time.tzinfo is None:
            start_time = start_time.astimezone(utc)
        if hasattr(self, 'data_type'):
            for key in ['strain rate', 'strain', 'displacement', 'velocity',
                        'acceleration']:
                if key in self.data_type.lower():
                    if self.data_type.lower() == key:
                        unit = key
                    else:
                        warnings.warn(f'Data type {self.data_type} is not '
                                      f'supported in lightguide. Set to {key}.')
                    break
                else:
                    warnings.warn(f'Data type {unit} is not supported in '
                                  'lightguide. Set to default (starin rate).')
                    unit = 'strain rate'
        else:
            print('Set unit to default (starin rate).')
            unit = 'strain rate'
        return Blast(self.data, start_time, self.fs,
                     start_channel=self.start_channel, channel_spacing=self.dx,
                     unit=unit)

    def save(self, fname=None, ftype=None, keep_format=False):
        """
        Save the instance as a pickle file or update the raw file and resave as
        new file.

        :param fname: str or pathlib.PosixPath. Path of new DAS data file to
            save.
        :param ftype: None or str. None for automatic detection), or 'pkl',
            'pickle', 'tdms', 'h5', 'hdf5', 'segy', 'sgy', 'npy'.
        :param keep_format: bool. If True, we will make a copy of the
            self.source file and make changes to it. This will strictly preserve
            the original format, but will cost more IO resources.
        """
        if fname is None:
            if hasattr(self, 'source'):
                fname_list = self.source.split('.')
                fname_list[-2] += '_new'
                fname = '.'.join(fname_list)
            else:
                fname = 'section.pkl'

        if ftype is None:
            ftype = str(fname).lower().split('.')[-1]

        for rtp in [('pickle', 'pkl'), ('hdf5', 'h5'), ('segy', 'sgy')]:
            ftype = ftype.replace(*rtp)

        if keep_format:
            if not hasattr(self, 'source'):
                raise ValueError('self.source not exit.')
            if not os.path.isfile(self.source):
                raise ValueError('self.source is not a file.')
            if ftype != self.source_type:
                raise ValueError('self.source_type is different from ftype.')
            write(self, fname, ftype=ftype, raw_fname=self.source)
        else:
            write(self, fname, ftype=ftype)

        return self

    def channel_data(self, use_channel, replace=False):
        """
        Extract data of one channel or several channels.
        """
        channel = np.array(use_channel).astype(int)
        channel -= self.start_channel
        data = self.data[channel]
        if replace:
            self.data = data
            self.start_channel += channel[0]
            self.start_distance += channel[0] * self.dx
            return self
        else:
            return data

    def plot(self, xmode='distance', tmode='origin', obj='waveform',
             kwargs_pro={}, **kwargs):
        """
        Plot several types of 2-D seismological data.

        :param xmode: str. 'distance' or 'channel'.
        :param tmode: str. 'origin', 'start', 'time' or 'sampling'. If
            origin_time is not defined, 'origin' and 'start' is the same.
        :param obj: str. Type of data to plot. It should be one of 'waveform',
            'phasepick', 'spectrum', 'spectrogram', 'fk', or 'dispersion'.
        :param kwargs_pro: dict. If obj is one of 'spectrum', 'spectrogram',
            'fk' and data is not specified, this parameter will be used to
            process the data to plot.
        :param ax: Matplotlib.axes.Axes or tuple. Axes to plot. A tuple for new
            figsize. If not specified, the function will directly display the
            image using matplotlib.pyplot.show().
        :param dpi: int. The resolution of the figure in dots-per-inch.
        :param title: str. The title of this axes.
        :param transpose: bool. Transpose the figure or not.
        :param cmap: str or Colormap. The Colormap instance or registered
            colormap name used to map scalar data to colors.
        :param vmin, vmax: Define the data range that the colormap covers.
        :param xlim, ylim: Set the x-axis and y-axis view limits.
        :param dB: bool. Transfer data unit to dB and take 1 as the reference
            value.
        :param xlog, ylog: bool. If True, set the x-axis' or y-axis' scale as
            log.
        :param xinv, yinv: bool. If True, invert x-axis or y-axis.
        :param xlabel, ylabel: bool or str. Whether to plot a label or what
            label to plot for x-axis or y-axis.
        :param xticklabels, yticklabels: bool or sequence of str. Whether to
            plot ticklabels or what ticklabels to plot for x-axis or y-axis.
        :param colorbar: bool, str or Matplotlib.axes.Axes. Bool means plot
            colorbar or not. Str means the location of colorbar. Axes means the
            Axes into which the colorbar will be drawn.
        :param t0, x0: The beginning of time and space. Use instance's
            properties by default
        :param pick: Sequence of picked phases. Required if obj=='phasepick'.
        :param c: Phase velocity sequence. Required if obj=='dispersion'.
        :param data: numpy.ndarray. Data to plot. Required if obj is not
            'spectrum', 'spectrogram' and 'fk'.
        :param f: Frequency sequence. Required if obj is one of 'spectrum',
            'spectrogram', 'fk' and data is specified, or obj is 'dispersion'.
        :param k: Wavenumber sequence. Required if obj=='fk' and data is
            specified.
        :param t: Time sequence. Required if obj=='spectrogram' and data is
            specified.
        :param savefig: str or bool. Figure name to save if needed. If True,
            it will be set to parameter obj.
        """
        if 'data' not in kwargs.keys():
            if obj == 'waveform' or obj == 'phasepick':
                data = deepcopy(self.data)
            elif obj == 'spectrum':
                data, f = self.spectrum(**kwargs_pro)
                kwargs['f'] = f
            elif obj == 'spectrogram':
                data, f, t = self.spectrogram(**kwargs_pro)
                kwargs['f'] = f
                kwargs['t'] = t
            elif obj == 'fk':
                data, f, k = self.fk_transform(**kwargs_pro)
                kwargs['f'] = f
                kwargs['k'] = k
            if hasattr(self, 'scale'):
                data *= self.scale
        else:
            data = kwargs.pop('data')

        if 'ax' not in kwargs.keys() and 'title' not in kwargs.keys():
            kwargs['title'] = obj

        if obj == 'phasepick':
            for phase, pck in kwargs['pick'].items():
                if len(pck) != 0:
                    pck = np.array(pck)
                    pck[:, 0] -= self.start_channel
                    kwargs['pick'][phase] = pck

        if xmode == 'channel':
            kwargs.setdefault('x0', self.start_channel)
        elif xmode == 'distance':
            kwargs.setdefault('x0', self.start_distance)
        if tmode in ['origin', 'start', 'time']:
            kwargs.setdefault('t0', self.start_time)
            if tmode == 'origin':
                if hasattr(self, 'origin_time'):
                    kwargs['t0'] -= self.origin_time
                    kwargs.setdefault('ylabel', 'Times(s) after occurance')
                else:
                    tmode == 'start'
            if tmode == 'start':
                kwargs['t0'] -= self.start_time
            tmode = 'time'
        if hasattr(self, 'data_type'):
            kwargs.setdefault('colorbar_label', self.data_type)

        plot(data, self.dx, self.fs, obj=obj, xmode=xmode, tmode=tmode,
             **kwargs)

    def rescaling(self, scale=None):
        """
        Scale data according to a scale factor.

        :param scale: float. It is required if the Section instance does
            not specify the attribute 'scale'.
        """
        if scale is None:
            try:
                self.data *= self['scale']
            except ValueError:
                print('Please specify a scale factor.')
        else:
            if hasattr(self, 'scale') and self.scale != scale:
                warnings.warn('The set scale is different from the previous '
                              'self.scale.')
            self.data *= scale
        self.scale = 1
        return self

    def phase2strain(self, lam, e, n, gl=None):
        """
        Convert the optical phase shift in radians to strain, or phase change
        rate to strain rate.

        :param lam: float. Operational optical wavelength in vacuum.
        :param e: float. photo-slastic scaling factor for logitudinal strain in
            isotropic material.
        :param n: float. Refractive index of the sensing fiber.
        :paran gl: float. Gauge length. Required if self.gauge_length has not
            been set.
        """
        if gl:
            self.gauge_length = gl
        self.data = phase2strain(self.data, lam, e, n, self.gauge_length)
        if hasattr(self, 'data_type'):
            if 'phase' not in self.data_type:
                warnings.warn('The data type is {}, not phase shift. But it' +
                              'still takes effect.'.format(self.data_type))
            else:
                self.data_type = self.data_type.replace(
                    'phase shift', 'strain')
                self.data_type = self.data_type.replace('phase change rate',
                                                        'strain rate')
        return self

    def normalization(self, method='z-score'):
        """
        Normalize for each individual channel using Z-score method.

        :param method: str. Method for normalization, should be one of 'max',
            'z-score', 'MAD' or 'one-bit'.
        """
        self.data = normalization(self.data, method=method)
        if hasattr(self, 'data_type'):
            self.data_type = 'normed ' + self.data_type
        return self

    def demeaning(self):
        """
        Demean signal by subtracted mean of each channel.
        """
        self.data = demeaning(self.data)
        return self

    def detrending(self):
        """
        Detrend signal by subtracted a linear least-squares fit to data.
        """
        self.data = detrending(self.data)
        return self

    def stacking(self, N, step=None):
        """
        Stack several channels to increase the signal-noise ratio(SNR).

        :param N: int. N adjacent channels stacked into 1.
        :param step: int. Interval of data stacking.
        """
        if step is None:
            step = N
        self.data = stacking(self.data, N, step=step)
        self.dx *= step
        if hasattr(self, 'gauge_length'):
            self.gauge_length += self.dx * (N - 1)
        return self

    def taper(self, *args, **kwargs):
        """
        See cosin_taper.
        """
        self.cosine_taper(*args, **kwargs)
        return self

    def cosine_taper(self, p=0.1, side='both'):
        """
        Taper using Tukey window.

        :param p: float or sequence of floats. Each float means decimal
            percentage of Tukey taper for corresponding dimension (ranging from
            0 to 1). Default is 0.1 which tapers 5% from the beginning and 5%
            from the end. If only one float is given, it only do for time
            dimension.
        :param side: str. 'both', 'left', or 'right'.
        """
        self.data = cosine_taper(self.data, p=p, side=side)
        return self

    def downsampling(self, xint=None, tint=None, stack=True,
                     lowpass_filter=True):
        """
        Downsample DAS data.

        :param xint: int. Spatial downsampling factor.
        :param tint: int. Time downsampling factor.
        :param stack: bool. If True, stacking will replace decimation.
        :param lowpass_filter: bool. Lowpass cheby2 filter before time
            downsampling or not.
        :return: Downsampled data.
        """
        self.data = downsampling(self.data, xint=xint, tint=tint, stack=stack,
                                 lowpass_filter=lowpass_filter)
        if xint and xint > 1:
            self.dx *= xint
            if hasattr(self, 'gauge_length'):
                self.gauge_length += self.dx * (xint - 1)
        if tint and tint > 1:
            self.fs /= tint
        return self

    def trimming(self, mode=1, xmin=None, xmax=None, tmin=None, tmax=None):
        """
        Cut data to given start and end distance/channel or time/sampling
        points.

        :param mode: int. 0 means the unit of boundary is channel number and
            sampling points; 1 means the unit of boundary is meters and seconds.
        :param xmin, xmax: int or float. Boundary of channel number (mode=0)
            or boundary of distance (mode=1).
            in meters.
        :param tmin, tmax: int, float or DASDateTime. Boundary of sampling
            points (mode=0) or boundary of time (mode=1). When mode=1,
            start_time is of type DASDateTime and tmin/tmax are floats,
            tmin/tmax means the time relative to start_time.
        """
        if mode == 1:
            if tmin is not None:
                try:
                    tmin = round((tmin - self.start_time) * self.fs)
                except TypeError:
                    tmin = round(tmin * self.fs)
                if tmin < 0:
                    warnings.warn('tmin is earlier than start_time. Set tmin '
                                  'to start_time.')
                    tmin = 0
                elif tmin >= self.nt:
                    raise ValueError('tmin is later than end_time.')
            else:
                tmin = 0

            if tmax is not None:
                try:
                    tmax = round((tmax - self.start_time) * self.fs)
                except TypeError:
                    tmax = round(tmax * self.fs)
                if tmax <= 0:
                    raise ValueError('tmax is earlier than start_time.')
                if tmax > self.nt:
                    warnings.warn('tmax is later than end_time. Set tmax to the'
                                  ' end_time.')
                    tmax = self.nt

            if xmin is not None:
                xmin = round((xmin - self.start_distance) / self.dx)
                if xmin < 0:
                    warnings.warn('xmin is smaller than start_distance. Set '
                                  'xmin to 0.')
                    xmin = 0
                elif xmin >= self.nch:
                    raise ValueError('xmin is later than end_distance.')
            else:
                xmin = 0

            if xmax is not None:
                xmax = round((xmax - self.start_distance) / self.dx)
                if xmax <= 0:
                    raise ValueError('xmax is smaller than start_distance.')
                if xmax > self.nch:
                    warnings.warn('xmax is later than end_distance. Set xmax '
                                  'to the array length.')
                    xmax = self.nch

        elif mode == 0:
            if tmin is None:
                tmin = 0
            else:
                tmin = int(tmin)
            if xmin is None:
                xmin = 0
            else:
                xmin = int(xmin - self.start_channel)
            if xmax is not None:
                xmax = int(xmax - self.start_channel)

        self.data = trimming(self.data, dx=self.dx, fs=self.fs, xmin=xmin,
                             xmax=xmax, tmin=tmin, tmax=tmax)

        self.start_time += tmin / self.fs
        self.start_distance += xmin * self.dx
        self.start_channel += xmin

        return self

    def padding(self, dn, reverse=False):
        """
        Pad DAS data with 0.

        :param dn: int or sequence of ints. Number of points to pad for both
            dimensions.
        :param reverse: bool. Set True to reverse the operation.
        """
        self.data = padding(self.data, dn, reverse=reverse)
        return self

    def _time_int_dif_attr(self, mode=0):
        for type_group in [['phase change rate', 'phase shift'],
                           ['strain rate', 'strain'],
                           ['acceleration', 'velocity', 'displacement']]:
            for (i, tp) in enumerate(type_group):
                if tp in self.data_type:
                    try:
                        self.data_type = self.data_type.replace(
                            tp, type_group[i + mode])
                    except BaseException:
                        operate = ('differentiate', 'integrate')[mode > 0]
                        print(f'Data type conversion error. Can not {operate} '
                              f'{self.data_type} data.')
                    return self
        warnings.warn('Unable to convert data type.')

    def time_integration(self, c=0):
        """
        Integrate DAS data in time.

        :param c: float. A constant added to the result.
        """
        self.data = time_integration(self.data, self.fs, c=c)
        if hasattr(self, 'data_type'):
            self._time_int_dif_attr(mode=1)
        return self

    def time_differential(self, prepend=0):
        """
        Differentiate DAS data in time.

        :param prepend: 'mean' or values to prepend to `data` along axis prior to
            performing the difference. 
        """
        self.data = time_differential(self.data, self.fs, prepend=prepend)
        if hasattr(self, 'data_type'):
            self._time_int_dif_attr(mode=-1)
        return self

    def distance_integration(self, c=0):
        """
        Differentiate DAS data in distance.

        :param c: float. A constant added to the result.
        """
        self.data = distance_integration(self.data, self.dx, c=c)
        self._strain2vel_attr()
        return self

    def bandpass(self, freqmin, freqmax, zi=None, **kwargs):
        """
        Filter data from 'freqmin' to 'freqmax' using Butterworth bandpass
        filter of 'corners' corners.

        :param freqmin: Pass band low corner frequency.
        :param freqmax: Pass band high corner frequency.
        :param zi : None, 0, or array_like. Initial conditions for the cascaded
            filter delays. It is a vector of shape (n_sections, nch, 2). Set to
            0 to trigger a output of the final filter delay values.
        :param corners: Filter corners / order.
        :param zerophase: If True, apply filter once forwards and once
            backwards. This results in twice the number of corners but zero
            phase shift in the resulting filtered data. Only valid when zi is
            None.
        """
        if zi is None:
            self.data = bandpass(self.data, self.fs, freqmin, freqmax, **kwargs)
            return self
        else:
            self.data, zf = bandpass(self.data, self.fs, freqmin, freqmax,
                                     zi=zi, **kwargs)
            return zf

    def bandstop(self, freqmin, freqmax, zi=None, **kwargs):
        """
        Filter data removing data between frequencies 'freqmin' and 'freqmax'
        using Butterworth bandstop filter of 'corners' corners.

        :param freqmin: Stop band low corner frequency.
        :param freqmax: Stop band high corner frequency.
        :param zi : None, 0, or array_like. Initial conditions for the cascaded
            filter delays. It is a vector of shape (n_sections, nch, 2). Set to
            0 to trigger a output of the final filter delay values.
        :param corners: Filter corners / order.
        :param zerophase: If True, apply filter once forwards and once
            backwards. This results in twice the number of corners but zero
            phase shift in the resulting filtered data. Only valid when zi is
            None.
        """

        if zi is None:
            self.data = bandstop(self.data, self.fs, freqmin, freqmax, **kwargs)
            return self
        else:
            self.data, zf = bandstop(self.data, self.fs, freqmin, freqmax,
                                     zi=zi, **kwargs)
            return zf

    def lowpass(self, freq, zi=None, **kwargs):
        """
        Filter data removing data over certain frequency 'freq' using
        Butterworth lowpass filter of 'corners' corners.

        :param freq: Filter corner frequency.
        :param zi : None, 0, or array_like. Initial conditions for the cascaded
            filter delays. It is a vector of shape (n_sections, nch, 2). Set to
            0 to trigger a output of the final filter delay values.
        :param corners: Filter corners / order.
        :param zerophase: If True, apply filter once forwards and once
            backwards. This results in twice the number of corners but zero
            phase shift in the resulting filtered data. Only valid when zi is
            None.
        """
        if zi is None:
            self.data = lowpass(self.data, self.fs, freq, **kwargs)
            return self
        else:
            self.data, zf = lowpass(self.data, self.fs, freq, zi=zi, **kwargs)
            return zf

    def lowpass_cheby_2(self, freq, **kwargs):
        """
        Filter data by passing data only below a certain frequency. The main
        purpose of this cheby2 filter is downsampling. This method will
        iteratively design a filter, whose pass band frequency is determined
        dynamically, such that the values above the stop band frequency are
        lower than -96dB.

        :param freq: The frequency above which signals are attenuated with 95
            dB.
        :param maxorder: Maximal order of the designed cheby2 filter.
        :param zi : None, 0, or array_like. Initial conditions for the cascaded
            filter delays. It is a vector of shape (n_sections, nch, 2). Set to 0 to
            trigger a output of the final filter delay values.
        :param ba: If True return only the filter coefficients (b, a) instead of
            filtering.
        :param freq_passband: If True return additionally to the filtered data,
            the iteratively determined pass band frequency.
        """
        output = lowpass_cheby_2(self.data, self.fs, freq, **kwargs)
        if isinstance(output, tuple):
            self.data = output[0]
            if len(output) == 2:
                return output[1]
            else:
                return output[1:]
        else:
            if kwargs.pop('ba', False):
                self.data = output
                return self
            else:
                return output

    def highpass(self, freq, zi=None, **kwargs):
        """
        Filter data removing data below certain frequency 'freq' using
        Butterworth highpass filter of 'corners' corners.

        :param freq: Filter corner frequency.
        :param zi : None, 0, or array_like. Initial conditions for the cascaded
            filter delays. It is a vector of shape (n_sections, nch, 2). Set to
            0 to trigger a output of the final filter delay values.
        :param corners: Filter corners / order.
        :param zerophase: If True, apply filter once forwards and once
            backwards. This results in twice the number of corners but zero
            phase shift in the resulting filtered data. Only valid when zi is
            None.
        """
        if zi is None:
            self.data = highpass(self.data, self.fs, freq, **kwargs)
            return self
        else:
            self.data, zf = highpass(self.data, self.fs, freq, zi=zi, **kwargs)
            return zf

    def envelope(self):
        """
        Computes the envelope of the given data. The envelope is determined by
        adding the squared amplitudes of the data and it's Hilbert-Transform and
        then taking the square-root. The envelope at the start/end should not be
        taken too seriously.
        """
        self.data = envelope(self.data)
        if hasattr(self, 'data_type'):
            self.data_type += ' envelope'
        return self

    def spectrum(self, taper=0.05, nfft='default'):
        """
        Computes the spectrum of the given data.

        :param taper: Decimal percentage of Tukey taper.
        :param nfft: Number of points for FFT. None = sampling points, 'default'
            = next power of 2 of sampling points.
        :return: Spectrum and frequency sequence.
        """
        return spectrum(self.data, self.fs, taper=taper, nfft=nfft)

    def spectrogram(self, **kwargs):
        """
        Computes the spectrogram of the given data.

        :param ch1, ch2, nch: int. Start channel, end channel and channel step
            for calculating the average spectrogram.
        :param nperseg: int. Length of each segment.
        :param noverlap: int. Number of points to overlap between segments. If
            None, noverlap = nperseg // 2.
        :param nfft: int. Length of the FFT used. None = nperseg.
        :param detrend : str or bool. Specifies whether and how to detrend each
            segment.  'linear' or 'detrend' or True = detrend, 'constant' or
            'demean' = demean.
        :param boundary: str or None. Specifies whether the input signal is
            extended at both ends, and how to generate the new values, in order
            to center the first windowed segment on the first input point. This
            has the benefit of enabling reconstruction of the first input point
            when the employed window function starts at zero. Valid options are
            ['even', 'odd', 'constant', 'zeros', None].
        :return: Spectrogram, frequency sequence and time sequence.
        """
        if 'ch1' in kwargs.keys():
            ch1 = int(kwargs.pop('ch1') - self.start_channel)
        else:
            ch1 = 0
        if 'ch2' in kwargs.keys():
            ch2 = int(kwargs.pop('ch2') - self.start_channel)
        else:
            ch2 = self.nch
        if 'nch' in kwargs.keys():
            nch = int(kwargs.pop('nch'))
        else:
            nch = 1

        return spectrogram(self.data[ch1:ch2:nch], self.fs, **kwargs)

    def fk_transform(self, **kwargs):
        """
        Transform the data to the fk domain using 2-D Fourier transform method

        :param taper: float or sequence of floats. Each float means decimal
            percentage of Tukey taper for corresponding dimension (ranging from
            0 to 1). Default is 0.1 which tapers 5% from the beginning and 5%
            from the end.
        :param nfft: Number of points for FFT. None means sampling points;
            'default' means next power of 2 of sampling points, which makes
            result smoother.
        """
        return fk_transform(self.data, self.dx, self.fs, **kwargs)

    def channel_checking(self, use=False, **kwargs):
        """
        Use the energy of each channel to determine which channels are bad.

        :param use: bool. If True, only keep the data of good channels in
            self.data and return self.
        :param deg: int. Degree of the fitting polynomial.
        :param thresh: int or float. The MAD multiple of bad channel energy
            lower than good channels.
        :param continuity: bool. Perform continuity checks on bad channels and
            good channels.
        :param adjacent: int. The number of nearby channels for continuity
            checks.
        :param toleration: int. The number of discontinuous channel allowed in
            each channel (including itself) in the continuity check.
        :param plot: bool or str. False means no plotting. Str or True means
            plotting while str gives a non-default filename.
        :return: self or Good channels and bad channels.
        """
        good_chn, bad_chn = channel_checking(self.data, **kwargs)
        good_chn = good_chn + self.start_channel
        bad_chn = bad_chn + self.start_channel
        if use:
            self.channel_data(good_chn, replace=True)
            return self
        else:
            return good_chn, bad_chn

    def turning_points(self, data_type='default', **kwargs):
        """
        Seek turning points in the DAS channel.

        :param data_type: str. If data_type is 'coordinate', data should include
            latitude and longitude (first two columns), and can also include
            depth (last column). If data_type is 'waveform', data should be
            continuous waveform, preferably containing signal with strong
            coherence (earthquake, traffic signal, etc.).
        :param thresh: For coordinate data, when the angle of the optical cables
            on both sides centered on a certain point exceeds thresh, it is
            considered an turning point. For waveform, thresh means the MAD
            multiple of adjacent channel cross-correlation values lower than
            their median.
        :param depth_info: bool. Optional if data_type is 'coordinate'. Whether
            depth (in meters) is included in the coordinate data and need to be
            used.
        :param channel_gap: int. Optional if data_type is 'coordinate'. The
            smaller the value is, the finer the segmentation will be. It is
            set to half the ratio of gauge length and channel interval by
            default.
        :return: list. Channel index of turning points.
        """
        if data_type == 'default':
            data_type = 'coordinate' if hasattr(self, 'geometry') else \
                'waveform'
        if data_type == 'coordinate':
            if hasattr(self, 'gauge_length'):
                kwargs.setdefault(
                    'channel_gap', self.gauge_length / self.dx / 2)
            if 'data' in kwargs.keys():
                output = turning_points(data_type=data_type, **kwargs)
            elif hasattr(self, 'geometry'):
                output = turning_points(self.geometry, data_type=data_type,
                                        **kwargs)
            else:
                raise ValueError('Geometry needs to be defined in DASdata, or '
                                 'coordinate data should be given.')
        else:
            output = turning_points(self.data, data_type=data_type, **kwargs)

        if isinstance(output, tuple):
            output = np.array(list(set(output[0]) | set(output[1])))
        else:
            output = np.array(output)
        output += self.start_channel
        self.turning_channels = output
        return output

    def spike_removal(self, nch=50, nsp=5, thresh=10):
        """
        Use a median filter to remove high-strain spikes in the data.

        :param nch: int. Number of channels over which to compute the median.
        :param nsp: int. Number of sampling points over which to compute the
            median.
        :param thresh: Ratio threshold over the median over which a number is
            considered to be an outlier.
        """
        self.data = spike_removal(self.data, nch=nch, nsp=nsp, thresh=thresh)
        return self

    def common_mode_noise_removal(self):
        """
        Remove common mode noise (sometimes called horizontal noise) from data.
        """
        self.data = common_mode_noise_removal(self.data)
        return self

    def curvelet_denoising(self, **kwargs):
        """
        Use curevelet transform to filter stochastic or/and cooherent noise.

        :param choice: int. 0 for Gaussian denoising using soft thresholding, 1
            for velocity filtering using the standard FK methodology and 2 for
            both.
        :param pad: float or sequence of floats. Each float means padding
            percentage before FFT for corresponding dimension. If set to 0.1
            will pad 5% before the beginning and after the end.
        :param noise: numpy.ndarray. Noise record as reference.
        :param soft_thresh: bool. True for soft thresholding and False for hard
            thresholding.
        :param vmin, vmax: float. Velocity range in m/s.
        :param flag: -1 choose only negative apparent velocities, 0 choose both
            postive and negative apparent velocities, 1 choose only positive
            apparent velocities.
        :param mode: str. 'remove' for denoising and 'retain' for decomposition.
        :param scale_begin: int. The beginning scale to do coherent denoising.
        :param nbscales: int. Number of scales including the coarsest wavelet
            level. Default set to ceil(log2(min(M,N)) - 3).
        :param nbangles: int. Number of angles at the 2nd coarsest level,
            minimum 8, must be a multiple of 4.
        """
        self.data = curvelet_denoising(self.data, dx=self.dx, fs=self.fs,
                                       **kwargs)
        return self

    def fk_filter(self, mode='retain', verbose=False, **kwargs):
        """
        Transform the data to the f-k domain using 2-D Fourier transform method,
        and transform back to the x-t domain after filtering.

        :param mode: str. 'remove' for denoising, 'retain' for extraction, and
            'decompose' for decomposition and not update self.data.
        :param verbose: If True, return filtered data, f-k spectrum,
            frequency sequence, wavenumber sequence and f-k mask.
        :param taper: float or sequence of floats. Each float means decimal
            percentage of Tukey taper for corresponding dimension (ranging from
            0 to 1). Default is 0.1 which tapers 5% from the beginning and 5%
            from the end.
        :param pad: Pad the data or not. It can be float or sequence of floats.
            Each float means padding percentage before FFT for corresponding
            dimension. If set to 0.1 will pad 5% before the beginning and after
            the end. 'default' means pad both dimensions to next power of 2.
            None or False means don't pad data before or during Fast Fourier
            Transform.
        :param fmin, fmax, kmin, kmax, vmin, vmax: float or or sequence of 2
            floats. Sequence of 2 floats represents the start and end of taper.
        :param edge: float. The width of fan mask taper edge.
        :param flag: -1 keep only negative apparent velocities, 0 keep both
            postive and negative apparent velocities, 1 keep only positive
            apparent velocities.
        """
        output = fk_filter(self.data, self.dx, self.fs, mode=mode,
                           verbose=verbose, **kwargs)
        if mode == 'decompose':
            sec1 = self.copy()
            sec2 = self.copy()
            sec1.data, sec2.data = output[:2]
            if verbose:
                return sec1, sec2, *output[2:]
            else:
                return sec1, sec2
        elif verbose:
            self.data = output[0]
            return output
        else:
            self.data = output
            return self

    def curvelet_windowing(self, mode='retain', **kwargs):
        """
        Use curevelet transform to keep cooherent signal with certain velocity
        range.

        :param mode: str. 'remove' for denoising, 'retain' for extraction, and
            'decompose' for decomposition and not update self.data.
        :param vmin, vmax: float. Velocity range in m/s.
        :param flag: -1 keep only negative apparent velocities, 0 keep both
            postive and negative apparent velocities, 1 keep only positive
            apparent velocities.
        :param pad: float or sequence of floats. Each float means padding
            percentage before FFT for corresponding dimension. If set to 0.1
            will pad 5% before the beginning and after the end.
        :param scale_begin: int. The beginning scale to do coherent denoising.
        :param nbscales: int. Number of scales including the coarsest wavelet
            level. Default set to ceil(log2(min(M,N)) - 3).
        :param nbangles: int. Number of angles at the 2nd coarsest level,
            minimum 8, must be a multiple of 4.
        """
        output = curvelet_windowing(self.data, self.dx, self.fs, mode=mode,
                                    **kwargs)
        if mode == 'decompose':
            sec1 = self.copy()
            sec2 = self.copy()
            sec1.data, sec2.data = output
            return sec1, sec2
        else:
            self.data = output
            return self

    def _strain2vel_attr(self):
        if hasattr(self, 'data_type'):
            if 'strain rate' in self.data_type:
                self.data_type = 'acceleration'
            elif 'strain' in self.data_type:
                self.data_type = 'velocity'
            else:
                warnings.warn(f'The data type is {self.data_type}, neither '
                              'strain nor strain rate. But it still takes '
                              'effect.')
        else:
            self.data_type = 'velocity or acceleration'
        return self

    def fk_rescaling(self, turning=None, verbose=False, **kwargs):
        """
        Convert strain / strain rate to velocity / acceleration by fk rescaling.

        :param turning: Sequence of int. Channel number of turning points. If
            self.turning exists, it will be used by default unless the parameter
            turning is set to False.
        :param verbose: If True and turning is not set, return f-k spectrum,
            frequency sequence, wavenumber sequence and f-k mask.
        :param taper: float or sequence of floats. Each float means decimal
            percentage of Tukey taper for corresponding dimension (ranging from
            0 to 1). Default is 0.1 which tapers 5% from the beginning and 5%
            from the end. If the turning parameter is set, this parameter will
            be invalid.
        :param pad: Pad the data or not. It can be float or sequence of floats.
            Each float means padding percentage before FFT for corresponding
            dimension. If set to 0.1 will pad 5% before the beginning and after
            the end. 'default' means pad both dimensions to next power of 2.
            None or False means don't pad data before or during Fast Fourier
            Transform.
        :param fmax, kmin, vmax: float or or sequence of 2 floats. Sequence of 2
            floats represents the start and end of taper. Setting these
            parameters can reduce artifacts.
        :param edge: float. The width of fan mask taper edge.
        """
        if hasattr(self, 'turning_channels') and turning is None:
            turning = np.array(self.turning_channels) - self.start_channel

        self._strain2vel_attr()
        if verbose and not turning:
            data_res, fk, f, k, mask = fk_rescaling(self.data, self.dx, self.fs,
                                                    verbose=True, **kwargs)
            self.data = data_res
            return fk, f, k, mask
        else:
            self.data = fk_rescaling(self.data, self.dx, self.fs, **kwargs)
            return self

    def curvelet_conversion(self, turning=None, **kwargs):
        """
        Use curevelet transform to convert strain/strain rate to
        velocity/acceleration.

        :param turning: Sequence of int. Channel number of turning points. If
            self.turning exists, it will be used by default unless the parameter
            turning is set to False.
        :param pad: float or sequence of floats. Each float means padding
            percentage before FFT for corresponding dimension. If set to 0.1
            will pad 5% before the beginning and after the end.
        :param scale_begin: int. The beginning scale to do conversion.
        :param nbscales: int. Number of scales including the coarsest wavelet
            level. Default set to ceil(log2(min(M,N)) - 3).
        :param nbangles: int. Number of angles at the 2nd coarsest level,
            minimum 8, must be a multiple of 4.
        """
        if hasattr(self, 'turning_channels') and turning is None:
            turning = np.array(self.turning_channels) - self.start_channel

        self.data = curvelet_conversion(self.data, self.dx, self.fs,
                                        turning=turning, **kwargs)
        self._strain2vel_attr()
        return self

    def slant_stacking(self, channel='all', turning=None, **kwargs):
        """
        Convert strain to velocity based on slant-stack.

        :param channel: int or list or 'all'. convert a certain channel number /
            certain channel range / all channels.
        :param turning: Sequence of int. Channel number of turning points. If
            self.turning exists, it will be used by default unless the parameter
            turning is set to False.
        :param L: int. the number of adjacent channels over which slowness is
            estimated.
        :param slm: float. Slowness x max
        :param sls: float. slowness step
        :param freqmin: Pass band low corner frequency.
        :param freqmax: Pass band high corner frequency.
        """
        if hasattr(self, 'turning_channels') and turning is None:
            turning = np.array(self.turning_channels) - self.start_channel

        if isinstance(channel, int):
            channel = [channel - self.start_channel]
        elif isinstance(channel, Iterable):
            channel = np.array(channel) - self.start_channel
        elif isinstance(channel, str) and channel == 'all':
            channel = list(range(self.nch))

        self.start_channel += channel[0]
        self.start_distance += channel[0] * self.dx
        self.data = slant_stacking(self.data, self.dx, self.fs, channel=channel,
                                   turning=turning, **kwargs)
        self._strain2vel_attr()
        return self


# Purpose: Several functions for analysis data quality and geometry of channels
# Author: Minzhe Hu, Zefeng Li
# Date: 2025.3.31
# Email: hmz2018@mail.ustc.edu.cn
import numpy as np
from copy import deepcopy
from geographiclib.geodesic import Geodesic
from pyproj import Proj


def robust_polyfit(data, deg, thresh):
    """
    Fit a curve with a robust weighted polynomial.

    :param data: 1-dimensional array.
    :param deg: int. Degree of the fitting polynomial
    :param thresh: int or float. Defined MAD multiple of outliers.
    :return: Fitting data
    """
    nch = len(data)
    channels = np.arange(nch)
    p_coef = np.polyfit(channels, data, deg)
    p_fit = np.poly1d(p_coef)
    old_data = p_fit(channels)
    mse = 1

    # robust fitting until the fitting curve changes < 0.1% at every point.
    while mse > 0.001:
        rsl = abs(data - old_data)
        mad = np.median(rsl)
        weights = np.zeros(nch)
        weights[rsl < thresh * mad] = 1
        p_coef = np.polyfit(channels, data, deg, w=weights)
        p_fit = np.poly1d(p_coef)
        new_data = p_fit(channels)
        mse = np.nanmax(np.abs((new_data - old_data) / old_data))
        old_data = new_data

    return new_data, weights


def _continuity_checking(lst1, lst2, adjacent=2, toleration=2):
    lst1_raw = deepcopy(lst1)
    for chn in lst1_raw:
        discont = [a for a in lst2 if abs(a - chn) <= adjacent]
        if len(discont) >= adjacent * 2 + 1 - toleration:
            lst1.remove(chn)
            lst2.append(chn)

    return lst1, lst2


def channel_checking(data, deg=10, thresh=5, continuity=True, adjacent=2,
                     toleration=2, mode='low', verbose=False):
    """
    Use the energy of each channel to determine which channels are bad.

    :param data: 2-dimensional np.ndarray. Axis 0 is channel number and axis 1
        is time series.
    :param deg: int. Degree of the fitting polynomial
    :param thresh: int or float. The MAD multiple of bad channel energy lower
        than good channels.
    :param continuity: bool. Perform continuity checks on bad channels and good
        channels.
    :param adjacent: int. The number of nearby channels for continuity checks.
    :param toleration: int. The number of discontinuous channel allowed in each
        channel (including itself) in the continuity check.
    :param mode: str. 'low' means bad channels have low amplitude, 'high' means
        bad channels have high amplitude, and 'both' means bad channels are
        likely to have low or high amplitude.
    :return: Good channels and bad channels.
    """
    nch = len(data)
    energy = np.log10(np.sum(data**2, axis=1))
    energy[energy == -np.inf] = -308

    # Remove abnormal value by robust polynomial fitting.
    fitted_energy, weights = robust_polyfit(energy, deg, thresh)
    deviation = energy - fitted_energy

    # Iterate eliminates outliers.
    mad = np.median(abs(deviation[weights > 0]))
    if mode == 'low':
        bad_chn = np.argwhere(deviation < -thresh * mad).ravel().tolist()
    elif mode == 'high':
        bad_chn = np.argwhere(deviation > thresh * mad).ravel().tolist()
    elif mode == 'both':
        bad_chn = np.argwhere(deviation < -thresh * mad).ravel().tolist() + \
                np.argwhere(deviation > thresh * mad).ravel().tolist()
    good_chn = list(set(range(nch)) - set(bad_chn))

    if continuity:
        # Discontinuous normal value are part of bad channels.
        good_chn, bad_chn = _continuity_checking(good_chn, bad_chn,
                                                 adjacent=adjacent,
                                                 toleration=toleration)

        # Discontinuous outliers are usually not bad channels.
        bad_chn, good_chn = _continuity_checking(bad_chn, good_chn,
                                                 adjacent=adjacent,
                                                 toleration=toleration)

    bad_chn = np.sort(np.array(bad_chn))
    good_chn = np.sort(np.array(good_chn))
    if verbose:
        return good_chn, bad_chn, energy, fitted_energy - thresh * mad

    return good_chn, bad_chn


def _channel_location(track_pt):
    track, tn = track_pt[:, :-1], track_pt[:, -1]
    dim = track.shape[1]
    l_track = np.sqrt(np.sum(np.diff(track, axis=0) ** 2, axis=1))
    l_track_cum = np.hstack(([0], np.cumsum(l_track)))
    idx_kp = np.where(tn >= 0)[0]

    interp_ch = []
    chn = np.floor(tn[idx_kp[0]]).astype(int)
    interp_ch.append([*track[idx_kp[0]], chn])
    if abs(chn - tn[idx_kp[0]]) > 1e-6:
        chn += 1

    seg_interval = []
    for i in range(1, len(idx_kp)):
        # calculate actual interval between known-channel points
        istart, iend = idx_kp[i - 1], idx_kp[i]
        n_chn_kp = tn[iend] - tn[istart]
        d_interp = (l_track_cum[iend] - l_track_cum[istart]) / n_chn_kp
        seg_interval.append([tn[istart], tn[iend], d_interp])

        l_res = 0  # remaining fiber length before counting the next segment
        # consider if the given channelnumber is not an integer
        chn_res = tn[istart] - int(tn[istart])
        if d_interp == 0:
            while chn < int(tn[iend]):
                chn += 1
                interp_ch.append([*track[istart, :], chn])
            continue
        for j in range(istart, iend):
            l_start = l_track[j] + l_res

            # if tp segment length is large for more than one interval, get the
            # channel loc
            if l_start >= d_interp * (1 - chn_res - 1e-6):
                # floor int, num of channel available
                n_chn_tp = int(l_start / d_interp + chn_res)
                l_new = (np.arange(n_chn_tp) + 1 - chn_res) * d_interp - \
                    l_res  # channel distance from segment start

                # interpolate the channel loc
                t_new = np.zeros((len(l_new), dim))
                for d in range(dim):
                    t_new[:, d] = np.interp(l_new, [0, l_track[j]],
                                            [track[j, d], track[j + 1, d]])

                # remaining length to add to next segment
                l_res = l_start - n_chn_tp * d_interp

                # write interpolated channel loc
                for ti in t_new:
                    chn += 1
                    interp_ch.append([*ti, chn])

                # handle floor int problem when l_start/d_interp is near an
                # interger
                if (d_interp - l_res) / d_interp < 1e-6:
                    chn += 1
                    interp_ch.append([*track[j + 1, :], int(tn[j + 1])])
                    l_res = 0
                chn_res = 0
            # if tp segment length is not enough for one interval, simply add
            # the length to next segment
            elif l_start < d_interp:
                l_res = l_start

    if abs(tn[iend] - int(tn[iend])) > 1e-6:
        chn += 1
        interp_ch.append([*track[iend, :], chn])

    return np.array(seg_interval), np.array(interp_ch)


def location_interpolation(known_pt, track_pt=None, dx=2, data_type='lonlat',
                           verbose=False):
    """
    Interpolate to obtain the positions of all channels.

    :param known_pt: np.ndarray. Points with known channel numbers. Each row
        includes 2 or 3 coordinates and a channel number.
    :param track_pt: np.ndarray. Optional fiber spatial track points without
        channel numbers. Each row includes 2 or 3 coordinates. Please ensure
        that the track points are arranged in increasing order of track number.
        If track points is not dense enough, please insert the coordinates of
        known points into track points in order.
    :param dx: Known points far from the track (> dx) will be excluded.
        Recommended setting is channel interval. The unit is m.
    :param data_type: str. Coordinate type. 'lonlat' ('lonlatheight') for
        longitude, latitude in degree (and height in meters), 'xy' ('xyz') for
        x, y (and z) in meters.
    :param verbose: bool. If True, return interpoleted channel location and
        segment interval.
    :return: Interpoleted channel location if verbose is False.
    """
    known_pt = known_pt[known_pt[:,-1].argsort()]
    dim = known_pt.shape[1] - 1
    if 'lonlat' in data_type:
        zone = np.floor((max(known_pt[:,0]) + min(known_pt[:,0])) / 2 / 6)\
            .astype(int) + 31
        DASProj = Proj(proj='utm', zone=zone, ellps='WGS84',
                       preserve_units=False)
        known_pt[:, 0], known_pt[:, 1] = DASProj(known_pt[:, 0], known_pt[:, 1])
    else:
        assert 'xy' in data_type, ('data_type should be \'lonlat\',\''
                                   'lonlatheight\', \'xy\' or \'xyz\'')

    if track_pt is None:
        seg_interval, interp_ch = _channel_location(known_pt)
    else:
        K = len(known_pt)
        T = len(track_pt)
        track_pt = np.c_[track_pt, np.zeros(T) - 1]
        if 'lonlat' in data_type:
            track_pt[:, 0], track_pt[:, 1] = DASProj(track_pt[:, 0],
                                                     track_pt[:, 1])

        # insert the known points into the fiber track data
        matrix = [np.tile(track_pt[:, d], (K, 1)) -
                  np.tile(known_pt[:, d], (T, 1)).T for d in range(dim)]

        dist = np.sqrt(np.sum(np.array(matrix) ** 2, axis=0))
        for k in range(K):
            if min(dist[k]) < dx:
                t_list = np.sort(np.where(dist[k] == min(dist[k]))[0])
                for t in t_list:
                    if track_pt[t, -1] == -1:
                        track_pt[t, -1] = known_pt[k, -1]
                        last_pt = t
                        break

        # interpolation with regular spacing along the fiber track
        try:
            track_pt = track_pt[:last_pt + 1]
        except NameError:
            print('All known points are too far away from the track points. If '
                  'they are reliable, they can be merged in sequence as track '
                  'points to input')
            return None

        seg_interval, interp_ch = _channel_location(track_pt)

    if data_type == 'lonlat':
        interp_ch[:, 0], interp_ch[:, 1] = \
            DASProj(interp_ch[:, 0], interp_ch[:, 1], inverse=True)

    if verbose:
        return interp_ch, seg_interval
    return interp_ch


def _xcorr(x, y):
    N = len(x)
    meanx = np.mean(x)
    meany = np.mean(y)
    stdx = np.std(np.asarray(x))
    stdy = np.std(np.asarray(y))
    c = np.sum((y - meany) * (x - meanx)) / (N * stdx * stdy)
    return c


def _horizontal_angle_change(geo, gap=10):
    nch = len(geo)
    angle = np.zeros(nch)
    for i in range(1, nch - 1):
        lon, lat = geo[i]
        lon_s, lat_s = geo[max(i - gap, 0)]
        lon_e, lat_e = geo[min(i + gap, nch - 1)]
        azi_s = Geodesic.WGS84.Inverse(lat_s, lon_s, lat, lon)['azi1']
        azi_e = Geodesic.WGS84.Inverse(lat, lon, lat_e, lon_e)['azi1']
        dazi = azi_e - azi_s
        if abs(dazi) > 180:
            dazi = -np.sign(dazi) * (360 - abs(dazi))
        angle[i] = dazi

    return angle


def _vertical_angle_change(geo, gap=10):
    nch = len(geo)
    angle = np.zeros(nch)
    for i in range(1, nch - 1):
        lon, lat, dep = geo[i]
        lon_s, lat_s, dep_s = geo[max(i - gap, 0)]
        lon_e, lat_e, dep_e = geo[min(i + gap, nch - 1)]
        s12_s = Geodesic.WGS84.Inverse(lat_s, lon_s, lat, lon)['s12']
        theta_s = np.arctan((dep - dep_s) / s12_s) / np.pi * 180
        s12_e = Geodesic.WGS84.Inverse(lat, lon, lat_e, lon_e)['s12']
        theta_e = np.arctan((dep_e - dep) / s12_e) / np.pi * 180
        angle[i] = theta_e - theta_s

    return angle


def _local_maximum_indexes(data, thresh):
    idx = np.where(data > thresh)[0]
    if len(idx):
        i = list(np.where(np.diff(idx) > 1)[0] + 1)
        if len(idx) - 1 not in i:
            i.append(len(idx) - 1)
        b = 0
        max_idx = []
        for e in i:
            max_idx.append(idx[b] + np.argmax(data[idx[b]:idx[e]]))
            b = e
        return max_idx
    else:
        return []


def turning_points(data, data_type='coordinate', thresh=5, depth_info=False,
                   channel_gap=3):
    """
    Seek turning points in the DAS channel.

    :param data: numpy.ndarray. Data used to seek turning points.
    :param data_type: str. If data_type is 'coordinate', data should include
        longitude and latitude (first two columns), and can also include depth
        (last column). If data_type is 'waveform', data should be continuous
        waveform, preferably containing signal with strong coherence
        (earthquake, traffic signal, etc.).
    :param thresh: For coordinate data, when the angle of the optical cables on
        both sides centered on a certain point exceeds thresh, it is considered
        an turning point. For waveform, thresh means the MAD multiple of
        adjacent channel cross-correlation values lower than their median.
    :param depth_info: bool. Optional if data_type is 'coordinate'. Whether
        depth (in meters) is included in the coordinate data and need to be
        used.
    :param channel_gap: int. Optional if data_type is 'coordinate'. The smaller
        the value is, the finer the segmentation will be. It is recommended to
        set it to half the ratio of gauge length and channel interval.
    :return: list. Channel index of turning points.
    """
    if data_type == 'coordinate':
        angle = _horizontal_angle_change(data[:, :2], gap=channel_gap)
        turning_h = _local_maximum_indexes(abs(angle), thresh)

        if depth_info:
            angle = _vertical_angle_change(data, gap=channel_gap)
            turning_v = _local_maximum_indexes(abs(angle), thresh)
            return turning_h, turning_v

        return turning_h

    elif data_type == 'waveform':
        nch = len(data)
        cc = np.zeros(nch - 1)
        for i in range(nch - 1):
            cc[i] = _xcorr(data[i], data[i + 1])
        median = np.median(cc)
        mad = np.median(abs(cc - median))

        return np.argwhere(cc < median - thresh * mad)[0]

    else:
        raise ValueError('Data_type should be \'coordinate\' or \'waveform\'.')


def _equally_spacing(dist, dx):
    index = [[], []]
    residual = [0, abs(dist[0]-dx)]
    for i in range(2, len(dist)+1):
        res = []
        for j in range(i):
            res.append(residual[j] + abs(dx - sum(dist[j:i])))
        residual.append(min(res))
        k = np.argmin(res)
        if k > 0:
            index.append(index[k] + [k])
        else:
            index.append(index[k])
        # print(index, residual)
    
    return index[-1]


def channel_spacing(geometry, depth_info=False):
    nch = len(geometry)
    dist = np.zeros(nch - 1)
    for i in range(nch - 1):
        lon0, lat0 = geometry[i, :2]
        lon1, lat1 = geometry[i+1, :2]
        d = Geodesic.WGS84.Inverse(lat0, lon0, lat1, lon1)['s12']
        if depth_info:
            dist[i] = np.sqrt(d**2 + (geometry[i+1, 2] - geometry[i, 2]) ** 2)
        else:
            dist[i] = d

    return dist


def distance_to_channels(geometry, points):
    """
    Calculate the distance from a point to each channel.

    :param geometry: numpy.ndarray. It needs to consist of two columns (
        longitude, latitude), three columns (longitude, latitude and depth).
    :param points: numpy.ndarray. A array consisting of longitude and
        latitude or longitude, latitude and depth.
    :return: numpy.ndarray. The distance from the given point to each channel.
    """
    if geometry.shape[1] == 3:
        depth_info = True
    else:
        depth_info = False

    nch = len(geometry)
    points = np.array(points)
    if points.ndim == 1:
        points = points.reshape(1, -1)
    npt = len(points)
    dist = np.zeros((npt, nch))
    for i, pt in enumerate(points):
        for j, geo in enumerate(geometry):
            d = Geodesic.WGS84.Inverse(pt[1], pt[0], geo[1], geo[0])['s12']
            if depth_info:
                dist[i, j] = np.sqrt(d**2 + (pt[2] - geo[2]) ** 2)
            else:
                dist[i, j] = d
    return dist


def closest_channel_to_point(geometry, points, verbose=False):
    """
    Find the channel number closest to a given point.

    :param geometry: numpy.ndarray. It needs to consist of longitude, latitude
        (and depth) or channel number, longitude, latitude (and depth).
    :param points: numpy.ndarray. A tuple consisting of longitude and
        latitude (and depth).
    :param verbose: bool. Return the channel and the distance to the closest
        channel if True.
    :return: int. The channel number closest to the given point.
    """
    nch = len(geometry)
    if points.shape[1] == geometry.shape[1]:
        channels = np.arange(nch).astype(int)
    else:
        geometry = geometry[geometry[:, 0].argsort()]
        channels = geometry[:, 0].astype(int)
        geometry = geometry[:, 1:]

    dist = distance_to_channels(points, geometry)
    closest_index = np.argmin(dist, axis=1)
    if verbose:
        return channels[closest_index], np.min(dist, axis=1)
    return channels[closest_index]


def equally_spaced_channels(geometry, dx, depth_info=False, verbose=False):
    """
    Find equally spaced channel numbers based on known DAS latitude and
    longitude.

    :param geometry: numpy.ndarray. DAS geometry used to filter equally spaced
        channels. It needs to consist of longitude, latitude (and depth) or
        channel number, longitude, latitude (and depth).
    :param dx: Channel interval.
    :param depth_info: bool. Whether depth (in meters) is included in the
        geometry and needed to be used.
    :param verbose: bool. If True, return channel numbers for equally spaced
        channels and channel intervals.
    :return: Channel numbers for equally spaced channels if verbose is False.
    """
    nch = len(geometry)
    if geometry.shape[1] == 2 + int(depth_info):
        channels = np.arange(nch).astype(int)
    else:
        geometry = geometry[geometry[:, 0].argsort()]
        channels = geometry[:, 0].astype(int)
        geometry = geometry[:, 1:]

    dist = channel_spacing(geometry, depth_info=False)

    s = 0
    idx_equal = [0]
    for i in range(nch-2):
        if dist[i] > dx * 2:
            e = i
            if e == s + 1:
                idx_equal.append(e)
            elif e >= s + 2:
                idx_equal.extend([idx + s for idx in
                                  _equally_spacing(dist[s:e], dx)])
                idx_equal.append(e)
            s = e + 1
            idx_equal.append(s)
        elif dist[i] + dist[i+1] > dx * 1.5:
            e = i + 1
            if e == s + 1:
                idx_equal.append(e)
            elif e >= s + 2:
                idx_equal.extend([idx + s for idx in
                                  _equally_spacing(dist[s:e], dx)])
                idx_equal.append(e)
            s = e
    e = nch - 1
    if e == s + 1:
        idx_equal.append(e)
    elif e >= s + 2:
        idx_equal.extend([idx + s for idx in
                        _equally_spacing(dist[s:e], dx)])
        idx_equal.append(e)

    return channels[idx_equal]